var prevContext = null;
var currentContext = null;

var getCurrentContext = function () { return currentContext; };
var setCurrentContext = function (ctx) {
  prevContext = currentContext;
  currentContext = ctx;
}
var revertContext = function () {
  currentContext = prevContext;
}

var createContext = function (label) {
  if ( label === void 0 ) label = 'anonymous';

  var ctx = {};
  var disposables = [];
  var hasRun = false;
  var state = {};

  var run = function (computation) {
    if (hasRun) {
      throw new Error('Each context can only run once');
    }

    hasRun = true;
    ctx.parent = currentContext;

    // inherit states
    if (ctx.parent) {
      var parentState = Object.create(ctx.parent.getState());
      state = Object.assign(parentState, state);
    }
    currentContext = ctx;

    computation();
    currentContext = ctx.parent;
  }

  // public API
  ctx.run = run;
  ctx.addDisposable = function (disposable) { return disposables.push(disposable); };
  ctx.dispose = function () { return disposables.forEach(function (fn) { return fn(); }); };
  ctx.getState = function () { return state; }
  ctx.toString = function () { return ("[Context " + label + "]"); };
  return ctx;
}

var withContext = function (fn) { return function () {
  var this$1 = this;
  var params = [], len = arguments.length;
  while ( len-- ) params[ len ] = arguments[ len ];

  var ctx = createContext();
  return ctx.run(function () {
    fn.apply(this$1, params);
  });
}; }

var native = {
  Timeout: global.setTimeout,
  Interval: global.setInterval,
}

var unpatch = function () {
  global.setTimeout = native.Timeout;
  global.setInteval = native.Interval;
};

var patch = function () {
  var methods = ['Timeout', 'Interval'];

  methods.forEach(function (method) { return global[("set" + method)] = function (fn, ms) {
      var params = [], len = arguments.length - 2;
      while ( len-- > 0 ) params[ len ] = arguments[ len + 2 ];

      var currentContext = getCurrentContext();
      if (!currentContext) {
        return native[method].apply(native, [ fn, ms ].concat( params ));
      }

      var computation = function () {
        setCurrentContext(currentContext);
        fn.apply(void 0, params);
        revertContext();
      }
      var id = native[method](computation, ms);
      var dispose = function () { return global[("clear" + method)](id); };
      currentContext.addDisposable(dispose);
      return id;
    }; }
  );

  return unpatch;
}

var methods = ['then', 'catch', 'finally'];
var natives = {};
var patched = false;

var unpatch$1 = function () {
  patched = false;
  methods.forEach(function (method) {
    global.Promise.prototype[method] = natives[method];
  })
};

var patch$1 = function () {
  if (patched) {
    return;
  }
  patched = true;
  var proto = Promise.prototype;
  methods.forEach(function (method) {
    // store the original version
    natives[method] = proto[method];

    // patching
    proto[method] = function () {
      var this$1 = this;
      var params = [], len = arguments.length;
      while ( len-- ) params[ len ] = arguments[ len ];

      var currentContext = getCurrentContext();
      if (!currentContext) {
        return natives[method].apply(this, params);
      }

      var wrappedParams = params.map(function (f) {
        if (typeof f !== 'function') {
          return f;
        }

        var computation = function () {
          var args = [], len = arguments.length;
          while ( len-- ) args[ len ] = arguments[ len ];

          setCurrentContext(currentContext);
          var ret = f.apply(this$1, args);
          revertContext();
          return ret;
        }

        return computation;
      });

      return natives[method].apply(this, wrappedParams);
    };
  });

  return unpatch$1;
}

patch();
if (typeof Promise !== 'undefined') {
  patch$1();
}

export { getCurrentContext, setCurrentContext, revertContext, createContext, withContext };